# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

import time
from timeit import default_timer as timer


class Runner:
    __added = []
    __removed = []
    __robots = []
    __thread = None
    __required = 0
    __checked = 0
    __start_flag = False
    __execute = None
    __evaluate = None
    __evaluate_result = False

    @staticmethod
    def shutdown():
        robots = Runner.__robots
        Runner.__robots = []
        for robot in robots:
            robot.dispose()

        Runner.__running = False
        thread = Runner.__thread
        Runner.__thread = None
        if thread:
            thread.join()

    @staticmethod
    def register_robot(robot):
        Runner.__added.append(robot)

    @staticmethod
    def unregister_robot(robot):
        Runner.__removed.append(robot)

    @staticmethod
    def register_required():
        Runner.__required += 1

    @staticmethod
    def register_checked():
        Runner.__checked += 1

    @staticmethod
    def set_executable(execute):
        Runner.__execute = execute

    @staticmethod
    def wait(milliseconds):
        if isinstance(milliseconds, (int, float)) and milliseconds > 0:
            timeout = timer() + milliseconds / 1000.0
            while timer() < timeout:
                time.sleep(0.000000001)

    @staticmethod
    def wait_until_ready():
        while Runner.__checked < Runner.__required:
            time.sleep(0.01)

    @staticmethod
    def wait_until(evaluate):
        Runner.__evaluate_result = False
        Runner.__evaluate = evaluate
        while Runner.__evaluate_result == False:
            time.sleep(0.01)

    @staticmethod
    def __run():
        try:
            target_time = timer()
            while Runner.__running:
                if timer() > target_time:
                    added = Runner.__added
                    removed = Runner.__removed
                    robots = Runner.__robots

                    if len(added) > 0:
                        for robot in added:
                            robots.append(robot)
                        Runner.__added = []
                    if len(removed) > 0:
                        for robot in removed:
                            robots.remove(robot)
                        Runner.__removed = []

                    for robot in robots:
                        robot._Robot__update_sensory_device_state()

                    if Runner.__evaluate:
                        result = False
                        try:
                            result = Runner.__evaluate.__func__()
                        except:
                            try:
                                result = Runner.__evaluate()
                            except:
                                Runner.__evaluate = None
                        Runner.__evaluate_result = result
                        if result:
                            Runner.__evaluate = None

                    if Runner.__execute:
                        try:
                            Runner.__execute.__func__()
                        except:
                            try:
                                Runner.__execute()
                            except:
                                Runner.__execute = None

                    for robot in robots:
                        robot._Robot__request_motoring_data()
                    for robot in robots:
                        robot._Robot__update_motoring_device_state()
                    for robot in robots:
                        robot._Robot__notify_motoring_device_data_changed()

                    target_time += 0.02
                time.sleep(0.000000001)
        except:
            pass

    @staticmethod
    def start():
        import threading
        if Runner.__start_flag == False:
            Runner.__start_flag = True
            Runner.__running = True
            thread = threading.Thread(target=Runner.__run)
            Runner.__thread = thread
            thread.daemon = True
            thread.start()